//----------------------------------------------------------------------------
// Anti-Grain Geometry - Version 2.3
// Copyright (C) 2002-2005 Maxim Shemanarev (http://www.antigrain.com)
//
// Permission to copy, use, modify, sell and distribute this software 
// is granted provided this copyright notice appears in all copies. 
// This software is provided "as is" without express or implied
// warranty, and with no claim as to its suitability for any purpose.
//
//----------------------------------------------------------------------------
//
// The author gratefully acknowleges the support of David Turner, 
// Robert Wilhelm, and Werner Lemberg - the authors of the FreeType 
// libray - in producing this work. See http://www.freetype.org for details.
//
//----------------------------------------------------------------------------
// Contact: mcseem@antigrain.com
//          mcseemagg@yahoo.com
//          http://www.antigrain.com
//----------------------------------------------------------------------------
//
// Adaptation for 32-bit screen coordinates has been sponsored by 
// Liberty Technology Systems, Inc., visit http://lib-sys.com
//
// Liberty Technology Systems, Inc. is the provider of
// PostScript and PDF technology for software developers.
// 
//----------------------------------------------------------------------------
#ifndef AGG_RASTERIZER_COMPOUND_AA_INCLUDED
#define AGG_RASTERIZER_COMPOUND_AA_INCLUDED

#include <string.h>
#include <math.h>
#include "agg_basics.h"
#include "agg_math.h"
#include "agg_array.h"
#include "agg_clip_liang_barsky.h"
#include "agg_render_scanlines.h"

#include "agg_rasterizer_scanline_aa.h"


namespace agg
{

    //-----------------------------------------------------------cell_style_aa
    // A pixel cell. There're no constructors defined and it was done 
    // intentionally in order to avoid extra overhead when allocating an 
    // array of cells.
    struct cell_style_aa
    {
        int   x;
        int   y;
        int16 left, right;
        int   cover;
        int   area;

        void set(int x, int y, int c, int a);
        void set(int x, int y, int l, int r, int c, int a);
        void set_coord(int x, int y);
        void set_cover(int c, int a);
        void add_cover(int c, int a);

        bool operator == (const cell_style_aa& c) const
        {
            return x    == c.x    && y     == c.y && 
                   left == c.left && right == c.right;
        }
        bool operator != (const cell_style_aa& c) const
        {
            return !(*this == c);
        }
    };


    //----------------------------------------------------outline_compound_aa
    // An internal class that implements the main rasterization algorithm.
    // Used in the rasterizer. Should not be used direcly.
    class outline_compound_aa
    {
        enum cell_block_scale_e
        {
            cell_block_shift = 12,
            cell_block_size  = 1 << cell_block_shift,
            cell_block_mask  = cell_block_size - 1,
            cell_block_pool  = 256,
            cell_block_limit = 1024
        };

        struct sorted_y
        {
            unsigned start;
            unsigned num;
        };

    public:
        ~outline_compound_aa();
        outline_compound_aa();

        void reset();

        void move_to(int x, int y, int l, int r);
        void line_to(int x, int y);

        int min_x()     const { return m_min_x; }
        int min_y()     const { return m_min_y; }
        int max_x()     const { return m_max_x; }
        int max_y()     const { return m_max_y; }
        int min_style() const { return m_min_style; }
        int max_style() const { return m_max_style; }

        void sort_cells();

        unsigned total_cells() const 
        {
            return m_num_cells;
        }

        unsigned scanline_num_cells(unsigned y) const 
        { 
            return m_sorted_y[y - m_min_y].num; 
        }

        const cell_style_aa* const* scanline_cells(unsigned y) const
        { 
            return m_sorted_cells.data() + m_sorted_y[y - m_min_y].start; 
        }

        bool sorted() const { return m_sorted; }

    private:
        outline_compound_aa(const outline_compound_aa&);
        const outline_compound_aa& operator = (const outline_compound_aa&);

        void set_cur_cell(int x, int y);
        void set_cur_cell(int x, int y, int r, int l);
        void add_cur_cell();
        void render_hline(int ey, int x1, int y1, int x2, int y2);
        void render_line(int x1, int y1, int x2, int y2);
        void allocate_block();
        
    private:
        unsigned m_num_blocks;
        unsigned m_max_blocks;
        unsigned m_cur_block;
        unsigned m_num_cells;
        cell_style_aa**           m_cells;
        cell_style_aa*            m_cur_cell_ptr;
        pod_array<cell_style_aa*> m_sorted_cells;
        pod_array<sorted_y>       m_sorted_y;
        cell_style_aa             m_cur_cell;
        int  m_cur_x;
        int  m_cur_y;
        int  m_min_x;
        int  m_min_y;
        int  m_max_x;
        int  m_max_y;
        int  m_min_style;
        int  m_max_style;
        bool m_sorted;
    };


    //------------------------------------------------------------------------
    template<unsigned XScale=1, unsigned AA_Shift=8> class rasterizer_compound_aa
    {
        enum status
        {
            status_initial,
            status_line_to,
            status_closed
        };

        struct style_info 
        { 
            int x, area, cover; 
        };

    public:
        enum aa_scale_e
        {
            aa_shift = AA_Shift,
            aa_num   = 1 << aa_shift,
            aa_mask  = aa_num - 1,
            aa_2num  = aa_num * 2,
            aa_2mask = aa_2num - 1
        };

        //--------------------------------------------------------------------
        rasterizer_compound_aa() : 
            m_clipped_start_x(0),
            m_clipped_start_y(0),
            m_style_left(-1),
            m_style_right(-1)
        {
        }


        //--------------------------------------------------------------------
        void reset(); 


        //--------------------------------------------------------------------
        void styles(int left, int right)
        {
            m_style_left = left;
            m_style_right = right;
        }

        void add_vertex(double x, double y, unsigned cmd);
        void move_to(int x, int y);
        void line_to(int x, int y);
        void move_to_d(double x, double y);
        void line_to_d(double x, double y);
        
        //--------------------------------------------------------------------
        int min_x() const { return m_outline.min_x(); }
        int min_y() const { return m_outline.min_y(); }
        int max_x() const { return m_outline.max_x(); }
        int max_y() const { return m_outline.max_y(); }

        //--------------------------------------------------------------------
        AGG_INLINE unsigned calculate_alpha(int area) const
        {
            int cover = area >> (poly_base_shift*2 + 1 - aa_shift);

            if(cover < 0) cover = -cover;

            //cover &= aa_2mask;
            //if(cover > aa_num)
            //{
            //    cover = aa_2num - cover;
            //}

            if(cover > aa_mask) cover = aa_mask;
            return cover;
        }

        //--------------------------------------------------------------------
        AGG_INLINE void sort()
        {
            m_outline.sort_cells();
        }

        //--------------------------------------------------------------------
        AGG_INLINE bool rewind_scanlines()
        {
            m_outline.sort_cells();
            if(m_outline.total_cells() == 0) 
            {
                return false;
            }
            if(m_outline.max_style() < m_outline.min_style())
            {
                return false;
            }
            m_cur_y = m_outline.min_y();
            return true;
        }

        //--------------------------------------------------------------------
        AGG_INLINE void add_style(unsigned style_id)
        {
            unsigned nbyte = style_id >> 3;
            unsigned mask = 1 << (style_id & 7);

            if((m_asm[nbyte] & mask) == 0)
            {
                m_ast.add(style_id);
                m_asm[nbyte] |= mask;
            }
        }

        //--------------------------------------------------------------------
        // Returns the number of styles
        unsigned sweep_styles()
        {
            for(;;)
            {
                if(m_cur_y > m_outline.max_y()) return 0;
                unsigned num_cells = m_outline.scanline_num_cells(m_cur_y);
                const cell_style_aa* const* cells = m_outline.scanline_cells(m_cur_y);
                unsigned num_styles = m_outline.max_style() - m_outline.min_style() + 1;

                m_ast.capacity(num_styles, 64);
                m_asm.allocate((num_styles + 7) >> 3, 8);
                m_asm.zero();

                while(num_cells--)
                {
                    const cell_style_aa* cur_cell = *cells++;
                    if(cur_cell->left  >= 0) add_style(cur_cell->left  - m_outline.min_style());
                    if(cur_cell->right >= 0) add_style(cur_cell->right - m_outline.min_style());
                }
                if(m_ast.size()) break;
                ++m_cur_y;
            }
            ++m_cur_y;
            return m_ast.size();
        }

        //--------------------------------------------------------------------
        // Returns style ID depending of the existing style index
        AGG_INLINE unsigned style(unsigned style_idx) const
        {
            return m_ast[style_idx] + m_outline.min_style();
        }


        //--------------------------------------------------------------------
        // Sweeps one scanline with one style index. The style ID can be 
        // determined by calling style(). 
        template<class Scanline> bool sweep_scanline(Scanline& sl, unsigned style_idx)
        {
            int cur_y = m_cur_y - 1;
            if(cur_y > m_outline.max_y()) return false;

            sl.reset_spans();

            int target_style = style(style_idx);

            unsigned num_cells = m_outline.scanline_num_cells(cur_y);
            const cell_style_aa* const* cells = m_outline.scanline_cells(cur_y);
            int cover = 0;

            while(num_cells)
            {
                int x    = 0;
                int area = 0;
                const cell_style_aa* cur_cell = 0;

                // Find the first cell that has the target style
                while(num_cells)
                {
                    cur_cell = *cells;
                    bool stop = false;
                    if(cur_cell->left == target_style)
                    {
                        x      = cur_cell->x;
                        area  += cur_cell->area;
                        cover += cur_cell->cover;
                        stop   = true;
                    }

                    if(cur_cell->right == target_style)
                    {
                        x      = cur_cell->x;
                        area  -= cur_cell->area;
                        cover -= cur_cell->cover;
                        stop   = true;
                    }
                    if(stop) break;
                    ++cells;
                    --num_cells;
                }


                if(num_cells)
                {
                    // Accumulate all cells with the same X
                    while(--num_cells)
                    {
                        cur_cell = *++cells;
                        if(cur_cell->x != x) break;

                        if(cur_cell->left == target_style)
                        {
                            area  += cur_cell->area;
                            cover += cur_cell->cover;
                        }
                        if(cur_cell->right == target_style)
                        {
                            area  -= cur_cell->area;
                            cover -= cur_cell->cover;
                        }
                    }
                }

                // Find the next cell with the appropriate target_style
                while(num_cells)
                {
                    if(cur_cell->left  == target_style ||
                       cur_cell->right == target_style) break;
                    --num_cells;
                    cur_cell = *++cells;
                }

                unsigned alpha;
                if(area)
                {
                    alpha = calculate_alpha((cover << (poly_base_shift + 1)) - area);
                    if(alpha)
                    {
                        sl.add_cell(x, alpha);
                    }
                    x++;
                }

                if(num_cells && cur_cell->x > x)
                {
                    alpha = calculate_alpha(cover << (poly_base_shift + 1));
                    if(alpha)
                    {
                        sl.add_span(x, cur_cell->x - x, alpha);
                    }
                }
            }
    
            if(sl.num_spans() == 0) return false;
            sl.finalize(cur_y);
            return true;
        }




        //--------------------------------------------------------------------
        void add_xy(const double* x, const double* y, unsigned n, int l, int r)
        {
            if(n > 2)
            {
                move_to_d(*x++, *y++);
                --n;
                do
                {
                    line_to_d(*x++, *y++);
                }
                while(--n);
            }
        }

        //-------------------------------------------------------------------
        template<class VertexSource>
        void add_path(VertexSource& vs, unsigned path_id=0)
        {
            double x;
            double y;

            unsigned cmd;
            vs.rewind(path_id);
            while(!is_stop(cmd = vs.vertex(&x, &y)))
            {
                add_vertex(x, y, cmd);
            }
        }


    private:
        //--------------------------------------------------------------------
        // Disable copying
        rasterizer_compound_aa(const rasterizer_compound_aa<XScale, AA_Shift>&);
        const rasterizer_compound_aa<XScale, AA_Shift>& 
            operator = (const rasterizer_compound_aa<XScale, AA_Shift>&);

    private:
        outline_compound_aa m_outline;
        pod_array<style_info> m_styles;
        pod_array<unsigned>   m_ast;      // Active Style Table
        pod_array<int8u>      m_asm;      // Active Style Mask

        int            m_clipped_start_x;
        int            m_clipped_start_y;
        int            m_style_left;
        int            m_style_right;
        int            m_cur_y;
    };










    //------------------------------------------------------------------------
    template<unsigned XScale, unsigned AA_Shift> 
    void rasterizer_compound_aa<XScale, AA_Shift>::reset() 
    { 
        m_outline.reset(); 
    }


    //------------------------------------------------------------------------
    template<unsigned XScale, unsigned AA_Shift> 
    void rasterizer_compound_aa<XScale, AA_Shift>::move_to(int x, int y)
    {
        if(m_outline.sorted()) 
        {
            reset();
        }
        m_outline.move_to(x * XScale, y, m_style_left, m_style_right); 
    }


    //------------------------------------------------------------------------
    template<unsigned XScale, unsigned AA_Shift> 
    void rasterizer_compound_aa<XScale, AA_Shift>::line_to(int x, int y)
    {
        m_outline.line_to(x * XScale, y); 
    }


    //------------------------------------------------------------------------
    template<unsigned XScale, unsigned AA_Shift> 
    void rasterizer_compound_aa<XScale, AA_Shift>::add_vertex(double x, double y, unsigned cmd)
    {
        if(is_move_to(cmd)) 
        {
            move_to(poly_coord(x), poly_coord(y));
        }
        else 
        {
            if(is_vertex(cmd))
            {
                line_to(poly_coord(x), poly_coord(y));
            }
        }
    }

    //------------------------------------------------------------------------
    template<unsigned XScale, unsigned AA_Shift> 
    void rasterizer_compound_aa<XScale, AA_Shift>::move_to_d(double x, double y) 
    { 
        move_to(poly_coord(x), poly_coord(y)); 
    }

    //------------------------------------------------------------------------
    template<unsigned XScale, unsigned AA_Shift> 
    void rasterizer_compound_aa<XScale, AA_Shift>::line_to_d(double x, double y) 
    { 
        line_to(poly_coord(x), poly_coord(y)); 
    }

}



#endif

