#include <math.h>
#include <stdio.h>
#include <time.h>
#include "agg_rendering_buffer.h"
#include "agg_conv_transform.h"
#include "agg_conv_stroke.h"
#include "agg_conv_clip_polyline.h"
#include "agg_scanline_u.h"
#include "agg_renderer_scanline.h"
#include "agg_rasterizer_outline_aa.h"
#include "agg_rasterizer_scanline_aa.h"
#include "agg_arc.h"
#include "agg_bezier_arc.h"
#include "agg_pixfmt_rgb.h"
#include "agg_bounding_rect.h"
#include "ctrl/agg_slider_ctrl.h"
#include "ctrl/agg_bezier_ctrl.h"
#include "ctrl/agg_rbox_ctrl.h"
#include "ctrl/agg_cbox_ctrl.h"
#include "platform/agg_platform_support.h"

#include "agg_rasterizer_compound_aa.h"

enum { flip_y = false };

typedef agg::pixfmt_bgr24 pixfmt;



namespace agg
{
    struct path_style
    {
        unsigned path_id;
        int left_fill;
        int right_fill;
        int line;
    };


    class compound_shape
    {
    public:
        ~compound_shape() 
        { 
            if(m_fd)
            {
                fclose(m_fd);
            }
        }

        compound_shape() :
            m_path(),
            m_affine(),
            m_curve(m_path),
            m_trans(m_curve, m_affine),
            m_styles()
        {}

        void open(const char* fname)
        {
            m_fd = fopen(fname, "r");
        }

        void read_next()
        {
            m_path.remove_all();
            m_styles.remove_all();
            const char space[] = " \t\n\r";
            double ax, ay, cx, cy;
            if(m_fd)
            {
                char buf[1024];
                char* ts;

                while(fgets(buf, 1022, m_fd))
                {
                    if(buf[0] == '=') break;
                }

                while(fgets(buf, 1022, m_fd))
                {
                    if(buf[0] == '!') break;
                    if(buf[0] == 'P')
                    {
                        // BeginPath
                        path_style style;
                        style.path_id = m_path.start_new_path();
                        ts = strtok(buf, space); // Path;
                        ts = strtok(0, space);  // left_style
                        style.left_fill = atoi(ts);
                        ts = strtok(0, space);  // right_style
                        style.right_fill = atoi(ts);
                        ts = strtok(0, space);  // line_style
                        style.line = atoi(ts);
                        ts = strtok(0, space);  // ax
                        ax = atof(ts);
                        ts = strtok(0, space);  // ay
                        ay = atof(ts);
                        m_path.move_to(ax, ay);
                        m_styles.add(style);
                    }


                    if(buf[0] == 'C')
                    {
                        ts = strtok(buf, space); // Curve;
                        ts = strtok(0, space);  // cx
                        cx = atof(ts);
                        ts = strtok(0, space);  // cy
                        cy = atof(ts);
                        ts = strtok(0, space);  // ax
                        ax = atof(ts);
                        ts = strtok(0, space);  // ay
                        ay = atof(ts);
                        m_path.curve3(cx, cy, ax, ay);
                    }

                    if(buf[0] == '<')
                    {
                        // EndPath
                    }
                }
            }
        }


        unsigned operator [] (unsigned i) const 
        {
            return m_styles[i].path_id;
        }


        unsigned paths() const { return m_styles.size(); }
        const path_style& style(unsigned i) const
        {
            return m_styles[i];
        }

        void rewind(unsigned path_id)
        {
            m_trans.rewind(path_id);
        }

        unsigned vertex(double* x, double* y)
        {
            return m_trans.vertex(x, y);
        }

        void scale(double w, double h)
        {
            m_affine.reset();
            double x1, y1, x2, y2;
            bounding_rect(m_path, *this, 0, m_styles.size(), 
                          &x1, &y1, &x2, &y2);
            if(x1 < x2 && y1 < y2)
            {
                double parl[6] = { 0, 0, w, 0, w, h };
                m_affine.rect_to_parl(x1, y1, x2, y2, parl);
            }
            m_curve.approximation_scale(m_affine.scale() * 10);
        }


    private:
        path_storage                              m_path;
        trans_affine                              m_affine;
        conv_curve<path_storage>                  m_curve;
        conv_transform<conv_curve<path_storage> > m_trans;
        pod_deque<path_style>                     m_styles;

        FILE* m_fd;
    };





    template<class Rasterizer, class Scanline, class Renderer, class Styles>
    void render_scanlines_compound_solid(Rasterizer& ras, 
                                         Scanline& sl, 
                                         Renderer& ren, 
                                         const Styles& st)
    {
        if(ras.rewind_scanlines())
        {
            sl.reset(ras.min_x(), ras.max_x());
            ren.prepare(unsigned(ras.max_x() - ras.min_x() + 2));

            unsigned num_styles;
            while((num_styles = ras.sweep_styles()) > 0)
            {
                unsigned i;
                for(i = 0; i < num_styles; i++)
                {
                    if(ras.sweep_scanline(sl, i))
                    {
                        ren.color(st[ras.style(i)]);
                        ren.render(sl);
                    }
                }
            }
        }
    }



}




class the_application : public agg::platform_support
{

public:
    typedef agg::renderer_base<pixfmt> renderer_base;
    typedef agg::renderer_scanline_aa_solid<renderer_base> renderer_scanline;
    typedef agg::rasterizer_scanline_aa<> rasterizer_scanline;
    typedef agg::scanline_u8 scanline;


    agg::compound_shape m_shape;
    agg::rgba8          m_colors[100];
    agg::trans_affine   m_scale;



    the_application(agg::pix_format_e format, bool flip_y) :
        agg::platform_support(format, flip_y)
    {
        for(unsigned i = 0; i < 100; i++)
        {

            m_colors[i] = agg::rgba8(
                (rand() & 0x7F) + 80, 
                (rand() & 0x7F) + 80, 
                (rand() & 0x7F) + 80, 
                150);

//            m_colors[i] = agg::rgba8(0, 0, 0, 128);

        }
    }


    void open(const char* fname)
    {
        m_shape.open(fname);
    }


    virtual void on_draw()
    {
        pixfmt pf(rbuf_window());
        renderer_base ren_base(pf);
        ren_base.clear(agg::rgba(1.0, 1.0, 0.95));
        renderer_scanline ren(ren_base);

        rasterizer_scanline ras;
        scanline sl;
        agg::conv_transform<agg::compound_shape> shape(m_shape, m_scale);
        agg::conv_stroke<agg::conv_transform<agg::compound_shape> > stroke(shape);

        agg::rasterizer_compound_aa<> rasc;

/*
        double dx = 0.5;
        double dy = 0.0;

        rasc.reset();
        rasc.styles(0, -1);
        rasc.move_to_d(100+dx, 100+dy);
        rasc.line_to_d(110+dx, 100+dy);
        rasc.line_to_d(110+dx, 110+dy);
        rasc.line_to_d(100+dx, 110+dy);
        rasc.line_to_d(100+dx, 100+dy);

        rasc.styles(1, -1);
        rasc.move_to_d(100+5+dx, 100+dy);
        rasc.line_to_d(110+5+dx, 100+dy);
        rasc.line_to_d(110+5+dx, 110+dy);
        rasc.line_to_d(100+5+dx, 110+dy);
        rasc.line_to_d(100+5+dx, 100+dy);

        agg::render_scanlines_compound(rasc, slc, renc, m_colors);
*/


/*
        double dx = 0.5;
        double dy = 0.0;

        rasc.reset();
        rasc.styles(0, -1);
        rasc.move_to_d(100+dx, 105+dy);
        rasc.line_to_d(100+dx, 100+dy);
        rasc.line_to_d(110+dx, 110+dy);
        rasc.line_to_d(100+dx, 120+dy);
        rasc.line_to_d(100+dx, 115+dy);


        rasc.styles(1, 0);
        rasc.move_to_d(100+dx, 105+dy);
        rasc.line_to_d(100+dx, 115+dy);


        rasc.styles(1, -1);
        rasc.move_to_d(100+dx, 115+dy);
        rasc.line_to_d(90+dx,  110+dy);
        rasc.line_to_d(100+dx, 105+dy);


        //rasc.styles(1, 0);
        //rasc.move_to_d(110, 102.4);
        //rasc.line_to_d(160, 150);
        //rasc.line_to_d(150, 190);
        //rasc.line_to_d(110, 102.4);

        agg::render_scanlines_compound(rasc, sl, ren, m_colors);
*/






        unsigned i;
        rasc.reset();
        for(i = 0; i < m_shape.paths(); i++)
        {
            
            if(m_shape.style(i).left_fill >= 0 || 
               m_shape.style(i).right_fill >= 0)
            {
                rasc.styles(m_shape.style(i).left_fill, 
                            m_shape.style(i).right_fill);
                rasc.add_path(shape, m_shape.style(i).path_id);
            }
        }
        agg::render_scanlines_compound_solid(rasc, sl, ren, m_colors);


        for(i = 0; i < m_shape.paths(); i++)
        {
            ras.reset();
            if(m_shape.style(i).line >= 0)
//            if(m_shape.style(i).left_fill >= 0 || 
//               m_shape.style(i).right_fill >= 0)
            {
                stroke.width(2);
                ras.add_path(stroke, m_shape.style(i).path_id);
                ren.color(agg::rgba8(0,0,0, 150));
                agg::render_scanlines(ras, sl, ren);
            }
           
        }




    }

    
    virtual void on_key(int x, int y, unsigned key, unsigned flags)
    {
        if(key == ' ')
        {
            m_shape.read_next();
            m_shape.scale(width(), height());
            force_redraw();
        }

        if(key == '+' || key == agg::key_kp_plus)
        {
            m_scale *= agg::trans_affine_translation(-x, -y);
            m_scale *= agg::trans_affine_scaling(1.1);
            m_scale *= agg::trans_affine_translation(x, y);
            force_redraw();
        }

        if(key == '-' || key == agg::key_kp_minus)
        {
            m_scale *= agg::trans_affine_translation(-x, -y);
            m_scale *= agg::trans_affine_scaling(1/1.1);
            m_scale *= agg::trans_affine_translation(x, y);
            force_redraw();
        }

    }

    virtual void on_ctrl_change()
    {
    }
};


int agg_main(int argc, char* argv[])
{
    the_application app(agg::pix_format_bgr24, flip_y);
    app.caption("AGG Example");
    const char* fname = "shapes";
    if(argc > 1) fname = argv[1];
    app.open(fname);
    if(app.init(655, 520, agg::window_resize))
    {
        return app.run();
    }
    return 1;
}





